from m3_gar.models.addrobj import *
from m3_gar.models.apartment import *
from m3_gar.models.carplace import *
from m3_gar.models.hierarchy import *
from m3_gar.models.history import *
from m3_gar.models.house import *
from m3_gar.models.normdoc import *
from m3_gar.models.op_type import *
from m3_gar.models.param import *
from m3_gar.models.reestr import *
from m3_gar.models.room import *
from m3_gar.models.stead import *
from m3_gar.models.version import *



# Если версия обновления на сайте ФИАС не содержит ссылки на файл с данными,
# то процесс обновления завершится ошибкой. Если в течение указанного срока с даты
# создания дампа версии ссылки не появилось, то эта версия будет пропущена
# и процесс обновления продолжится
VERSION_WITHOUT_URL_TIMEOUT_HOURS = 7 * 24  # неделя



from django.db import (
    models,
)
from django.db.models import (
    BooleanField,
    Case,
    Value,
    When,
)

from m3_gar.enums import (
    VersionUpdateStatus,
)


__all__ = ['Version']


class VersionManager(models.Manager):

    def get_queryset(self):
        query = super().get_queryset()

        query = query.annotate(
            processed=Case(
                When(
                    status__in=VersionUpdateStatus.get_processed_statuses(),
                    then=Value(True)
                ),
                default=Value(False),
                output_field=BooleanField(),
            )
        )

        return query

    def nearest_by_date(self, date):
        return self.get_queryset().filter(dumpdate__lte=date).latest('dumpdate')

    def last_processed(self):
        """Возвращает самую последнюю обработанную версию."""
        return self.get_queryset().filter(
            processed=True,
        ).last()


class Version(models.Model):

    objects = VersionManager()

    ver = models.IntegerField(primary_key=True)
    date = models.DateField(db_index=True, blank=True, null=True)
    dumpdate = models.DateField(db_index=True)

    complete_xml_url = models.CharField(max_length=255)
    delta_xml_url = models.CharField(max_length=255, blank=True, null=True)

    status = models.CharField(
        'Статус обновления',
        max_length=20,
        choices=VersionUpdateStatus.choices,
        default=VersionUpdateStatus.NEW,
    )

    def __str__(self):
        return f'{self.ver} от {self.dumpdate}'



from m3_gar.base_models import (
    Steads as BaseSteads,
)
from m3_gar.models.op_type import (
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['Steads']


class Steads(BaseSteads, RegionCodeModelMixin):
    """
    Сведения по земельным участкам
    """

    level = 9

    class Meta:
        verbose_name = 'Земельный участок'
        verbose_name_plural = 'Земельные участки'


make_fk(Steads, 'opertypeid', to=OperationTypes)
make_fk(Steads, 'objectid', to=ReestrObjects)

add_params(Steads, 'm3_gar.SteadsParams')



# Generated by Django 3.2.6 on 2022-02-17 05:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0006_auto_20211105_0937'),
    ]

    operations = [
        migrations.AddField(
            model_name='version',
            name='processed',
            field=models.BooleanField(default=False, verbose_name='Обработано'),
        ),
    ]



import os
from argparse import (
    BooleanOptionalAction,
)
from typing import (
    Optional,
)

from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
    CommandError,
)
from django.utils.translation import (
    activate,
)

from m3_gar.importer.commands import (
    auto_update_data,
    load_complete_data,
)
from m3_gar.importer.consts import (
    DEFAULT_BULK_LIMIT,
)
from m3_gar.importer.exceptions import (
    ImporterError,
)
from m3_gar.importer.timer import (
    Timer,
)
from m3_gar.importer.version import (
    fetch_version_info,
)
from m3_gar.models import (
    Version,
)
from m3_gar.util import (
    get_table_names_from_models,
)


class Command(BaseCommand):
    help = 'Fill or update GAR database'

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--src',
            help=(
                "Directory or archive path or url to load into DB. "
                "Use 'auto' to load latest known version"
            ),
        )
        parser.add_argument(
            '--truncate',
            action=BooleanOptionalAction,
            default=None,
            help='Truncate tables before loading data',
        )
        parser.add_argument(
            '--no-transaction',
            action='store_true',
            default=False,
            help='Do not wrap import in transaction',
        )
        parser.add_argument(
            '--update',
            action='store_true',
            help="Update database from https://fias.nalog.ru",
        )
        parser.add_argument(
            '--limit',
            type=int,
            default=DEFAULT_BULK_LIMIT,
            help="Limit rows for bulk operations",
        )
        parser.add_argument(
            '--tables',
            help="Comma-separated list of tables to import",
        )
        parser.add_argument(
            '--update-version-info',
            action=BooleanOptionalAction,
            default=True,
            help='Update list of available database versions from http://fias.nalog.ru',
        )
        parser.add_argument(
            '--tempdir',
            help="Path to the temporary files directory"
        )
        parser.add_argument(
            '--skip-versions',
            type=int,
            nargs='*',
            help='Номера версий обновлений, которые не будут выполняться',
        )

    def handle(
        self, *args,
        src,
        truncate,
        no_transaction,
        update,
        limit,
        tables,
        update_version_info,
        tempdir,
        skip_versions,
        **options,
    ):
        Timer.init()

        # признак обновления из внешнего источника
        remote = False
        if src and src.lower() == 'auto':
            src = None
            remote = True

        tempdir = self.parse_tempdir_arg(tempdir)

        if (src or remote) and Version.objects.filter(processed=True).exists() and truncate is None:
            self.stderr.write(
                'One of the tables contains data. '
                'Truncate all GAR tables manually or use '
                '--truncate/--no-truncate option'
            )

            raise CommandError

        if update_version_info:
            fetch_version_info(update_all=True)

        # Force Russian language for internationalized projects
        if settings.USE_I18N:
            activate('ru')

        tables = self.parse_tables_arg(tables)

        if src or remote:
            load_complete_data(
                path=src,
                truncate=truncate,
                no_transaction=no_transaction,
                limit=limit,
                tables=tables,
                tempdir=tempdir,
            )

        if update:
            try:
                auto_update_data(
                    limit=limit,
                    tables=tables,
                    tempdir=tempdir,
                    skip_versions=skip_versions,
                )
            except ImporterError as e:
                self.stdout.write(str(e))

    def parse_tempdir_arg(self, tempdir):
        """
        Возвращает временную директорую
        """
        if tempdir:
            error = None

            if not os.path.exists(tempdir):
                error = f'Directory `{tempdir}` does not exists.'
            elif not os.path.isdir(tempdir):
                error = f'Path `{tempdir}` is not a directory.'
            elif not os.access(tempdir, os.W_OK):
                error = f'Directory `{tempdir}` is not writeable'

            if error:
                self.stderr.write(error)

                raise CommandError

        return tempdir

    def parse_tables_arg(self, tables: Optional[str]) -> tuple[str, ...]:
        """
        Возвращает перечень таблиц для загрузки.
        """
        tables = set(tables.split(',')) if tables else set()
        tables_from_db = get_table_names_from_models()

        if not tables.issubset(set(tables_from_db)):
            diff = ', '.join(tables.difference(tables_from_db))
            self.stderr.write(
                f'Tables `{diff}` are not of GAR schema models and can not be processed'
            )

            raise CommandError

        tables = [x for x in tables_from_db if x in tables]

        return tuple(tables)



# Generated by Django 3.2.4 on 2025-04-14 10:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0022_auto_20250331_0342'),
    ]

    operations = [
        migrations.AlterField(
            model_name='operationtypes',
            name='name',
            field=models.CharField(max_length=300, verbose_name='Наименование'),
        ),
    ]



from m3_gar.base_models import (
    Carplaces as BaseCarplaces,
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['Carplaces']


class Carplaces(BaseCarplaces, RegionCodeModelMixin):
    """
    Сведения по машино-местам
    """

    level = 17

    class Meta:
        verbose_name = 'Машино-место'
        verbose_name_plural = 'Машино-места'


make_fk(Carplaces, 'opertypeid', to=OperationTypes)
make_fk(Carplaces, 'objectid', to=ReestrObjects)

add_params(Carplaces, 'm3_gar.CarplacesParams')



from django.contrib.postgres.indexes import (
    HashIndex,
)

from m3_gar.base_models import (
    Param as BaseParam,
    ParamTypes as BaseParamTypes,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    make_fk,
)


__all__ = [
    'ParamTypes',
    'AddrObjParams',
    'ApartmentsParams',
    'CarplacesParams',
    'HousesParams',
    'RoomsParams',
    'SteadsParams',
]


class ParamTypes(BaseParamTypes):
    """
    Сведения по типу параметра
    """
    class Meta:
        verbose_name = 'Тип параметра'
        verbose_name_plural = 'Типы параметров'


class Param(BaseParam, RegionCodeModelMixin):
    """
    Сведения о классификаторе параметров адресообразующих элементов и объектов недвижимости
    """
    class Meta:
        abstract = True
        verbose_name = 'Параметр'
        verbose_name_plural = 'Параметры'

        indexes = (
            HashIndex(
                fields=('objectid',),
            ),
        )


make_fk(Param, 'typeid', to=ParamTypes)


class AddrObjParams(Param):
    pass


class ApartmentsParams(Param):
    pass


class CarplacesParams(Param):
    pass


class HousesParams(Param):
    pass


class RoomsParams(Param):
    pass


class SteadsParams(Param):
    pass



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0021_remove_version_processed'),
    ]

    operations = [
        migrations.CreateModel(
            name='AddrCacheResult',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, verbose_name='Наименование')),
                ('page', models.CharField(max_length=10, verbose_name='Номер страницы')),
                ('data', models.TextField(verbose_name='Итоговые строки')),
            ],
            options={
                'verbose_name': 'Готовые результаты по запросу конкретного адреса',
                'verbose_name_plural': 'Готовые результаты по запросу конкретного адреса',
            },
        ),
        migrations.AddIndex(
            model_name='addrcacheresult',
            index=models.Index(fields=['name'], name='m3_gar_addr_name_ec91da_idx'),
        ),
    ]



import re
from pathlib import (
    PurePath,
)

from m3_gar.importer.table.exceptions import (
    BadTableError,
)
from m3_gar.importer.table.xml import (
    XMLTable,
)


table_xml_re = re.compile(
    r'as_(?P<name>[a-z_]+)_(?P<date>\d{8})_(?P<uuid>[a-z0-9-]{36}).xml',
    re.I,
)


class TableFactory:

    @staticmethod
    def parse(filename):
        path_obj = PurePath(filename)
        basename = path_obj.name

        if path_obj.parent == PurePath():
            region = None
        else:
            region = int(path_obj.parent.name)

        m = table_xml_re.match(basename)
        if m is not None:
            return XMLTable(filename=filename, region=region, **m.groupdict())

        return None



import datetime

from django import (
    db,
)
from django.conf import (
    settings,
)
from progress import (
    Infinite,
)

from m3_gar.importer.consts import (
    DEFAULT_BULK_LIMIT,
)
from m3_gar.importer.signals import (
    post_import_table,
    pre_import_table,
)


class LoadingBar(Infinite):
    """
    Индикатор загрузки
    T: Table (Таблица) - импортируемая в данный момент таблица
    L: Loaded (Загружено) - количество уже загруженных в таблицу строк
    U: Updated (Обновлено) - количество обновлённых записей
    FN: Filename (Имя файла) - имя файла импортируемой таблицы
    """

    text = (
        'T: %(table)s.'
        ' L: %(loaded)d'
        ' | U: %(updated)d'
        ' \tFN: %(filename)s'
    )

    hide_cursor = False
    check_tty = False

    def __init__(self, message=None, **kwargs):
        self.loaded = 0
        self.updated = 0
        self.table = kwargs.pop('table', 'unknown')
        self.filename = kwargs.pop('filename', 'unknown')

        super().__init__(message=message, **kwargs)

    def __getitem__(self, key):
        if key.startswith('_'):
            return None
        return getattr(self, key, None)

    def update(self, loaded=0, updated=0):
        if loaded:
            self.loaded = loaded
        if updated:
            self.updated = updated

        ln = self.text % self
        self.writeln(ln)


class TableLoader:
    """
    Класс для сохранения новых данных в БД. Обеспечивает пакетное сохранение
    согласно указанному лимиту и отображение индикатора загрузки в консоли.
    Обновление не поддерживается. При попытке сохранения данных по уже
    существующему в БД ключу может быть выброшен IntegrityError
    """

    def __init__(self, limit=DEFAULT_BULK_LIMIT):
        self.limit = int(limit)
        self.counter = 0
        self.upd_counter = 0
        self.today = datetime.date.today()

    def create(self, table, objects, bar):
        table.model.objects.bulk_create(objects)

        if settings.DEBUG:
            db.reset_queries()

    def load(self, tablelist, table):
        pre_import_table.send(sender=self.__class__, table=table)
        self.do_load(tablelist=tablelist, table=table)
        post_import_table.send(sender=self.__class__, table=table)

    def do_load(self, tablelist, table):
        bar = LoadingBar(table=table.name, filename=table.filename)
        bar.update()

        objects = []
        for item in table.rows(tablelist=tablelist):
            objects.append(item)
            self.counter += 1

            if self.counter and self.counter % self.limit == 0:
                self.create(table, objects, bar=bar)
                objects = []
                bar.update(loaded=self.counter)

        if objects:
            self.create(table, objects, bar=bar)

        bar.update(loaded=self.counter)
        bar.finish()


class TableUpdater(TableLoader):
    """
    Класс для сохранения или обновления данных в БД.
    Субкласс TableLoader, отличается поддержкой обновления данных.
    """

    def __init__(self, limit=DEFAULT_BULK_LIMIT):
        self.upd_limit = 100

        super().__init__(limit=limit)

    def do_load(self, tablelist, table):
        bar = LoadingBar(table=table.name, filename=table.filename)

        model = table.model
        objects = []
        for item in table.rows(tablelist=tablelist):
            try:
                model.objects.only('pk').get(pk=item.pk)
            except model.DoesNotExist:
                objects.append(item)
                self.counter += 1
            else:
                item.save()
                self.upd_counter += 1

            if self.counter and self.counter % self.limit == 0:
                self.create(table, objects, bar=bar)
                objects = []
                bar.update(loaded=self.counter)

            if self.upd_counter and self.upd_counter % self.upd_limit == 0:
                bar.update(updated=self.upd_counter)

        if objects:
            self.create(table, objects, bar=bar)

        bar.update(loaded=self.counter, updated=self.upd_counter)
        bar.finish()



from django.contrib.postgres.indexes import (
    BTreeIndex,
    HashIndex,
)
from django.db import (
    models,
)
from django.utils.functional import (
    cached_property,
)
from m3_gar.base_models import (
    AddrObj as BaseAddrObj,
    AddrObjDivision as BaseAddrObjDivision,
    AddrObjTypes as BaseAddrObjTypes,
)
from m3_gar.models.indexes import (
    UpperGinIndex,
)
from m3_gar.models.op_type import (
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['AddrObj', 'AddrObjDivision', 'AddrObjTypes', 'AddrCacheResult']


class AddrObjDivision(BaseAddrObjDivision, RegionCodeModelMixin):
    """Сведения по операциям переподчинения."""

    class Meta:
        verbose_name = 'Операция переподчинения'
        verbose_name_plural = 'Операции переподчинения'


class AddrObjTypes(BaseAddrObjTypes):
    """Сведения по типам адресных объектов."""

    class Meta:
        verbose_name = 'Тип адресного объекта'
        verbose_name_plural = 'Типы адресных объектов'

    @property
    def is_prefix(self):
        """Признак того, является ли тип объекта префиксом."""

        prefix_short_names = {
            'г', 'г.', 'г.ф.з.', 'Респ', 'респ.', 'г.о.', 'г', 'г.', 'дп', 'кп', 'п', 'пгт', 'пгт.', 'рп',
            'автодорога', 'гп', 'д', 'дп.', 'остров', 'с', 'с.',
        }

        return self.shortname in prefix_short_names


class AddrObj(BaseAddrObj, RegionCodeModelMixin):
    """Сведения классификатора адресообразующих элементов."""

    class Meta:
        verbose_name = 'Адресообразующий элемент'
        verbose_name_plural = 'Адресообразующие элементы'

        indexes = (
            HashIndex(
                fields=('objectguid',),
            ),
            BTreeIndex(
                fields=('isactive', 'isactual', 'level'),
            ),
            BTreeIndex(
                fields=('level',),
            ),
            UpperGinIndex(
                fields=['name_with_typename'],
                name='addrobj_name_with_typename_gin',
                opclasses=['gin_trgm_ops'],
            ),
        )

    @cached_property
    def addr_obj_type(self):
        """Тип адресного объекта."""

        try:
            result = AddrObjTypes.objects.get(
                shortname=self.typename,
                level=self.level,
            )
        except AddrObjTypes.DoesNotExist:
            result = AddrObjTypes()

        return result

    @property
    def type_full_name(self):
        """Полное наименование типа."""

        return self.addr_obj_type.name

    @property
    def is_prefix_type(self):
        """Признак того, является ли тип объекта префиксом."""

        # 1 - федеральный уровень адресных объектов
        return self.level != '1' or self.addr_obj_type.is_prefix


class AddrCacheResult(models.Model):
    """Готовые результаты по запросу конкретного адреса"""

    name = models.CharField(
        'Наименование',
        max_length=30,
    )
    page = models.CharField(
        'Номер страницы',
        max_length=10,
    )
    data = models.TextField(
        'Итоговые строки',
    )

    class Meta:
        verbose_name = 'Готовые результаты по запросу конкретного адреса'
        verbose_name_plural = 'Готовые результаты по запросу конкретного адреса'
        indexes = [
            models.Index(
                fields=['name'],
            ),
        ]


# На момент описания моделей AddrObjTypes никак не связывается с AddrObj
# Существующее поле AddrObj.typename - текстовое представление (ул, пер, г, и т.п.)
# make_fk(AddrObj, '???', to=AddrObjTypes, null=True, blank=True)

make_fk(AddrObjDivision, 'parentid', to=ReestrObjects, db_constraint=False)
make_fk(AddrObjDivision, 'childid', to=ReestrObjects, db_constraint=False)

make_fk(AddrObj, 'opertypeid', to=OperationTypes)
make_fk(AddrObj, 'objectid', to=ReestrObjects)

add_params(AddrObj, 'm3_gar.AddrObjParams')



from django.db.models import (
    TextChoices,
)


class VersionUpdateStatus(TextChoices):
    """Статус обработки версии обновления."""

    NEW = 'new', 'Новое'
    SKIPPED = 'skipped', 'Пропущено'
    ERROR = 'error', 'Ошибка'
    FINISHED = 'finished', 'Завершено'

    @classmethod
    def get_processed_statuses(cls) -> set['VersionUpdateStatus']:
        """Статусы обновлений, которые считаются обработанными."""
        return {cls.FINISHED, cls.SKIPPED}



from django.db import (
    models,
)

from m3_gar.base_models import (
    AdmHierarchy as BaseAdmHierarchy,
    MunHierarchy as BaseMunHierarchy,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    HierarchyMixin,
    RegionCodeModelMixin,
    make_fk,
)
from m3_gar.models.indexes import (
    UpperGinIndex,
)


__all__ = ['AdmHierarchy', 'MunHierarchy']


class Hierarchy(HierarchyMixin):
    """
    Базовый тип сведений по иерархии
    """

    name_with_parents = models.CharField(max_length=500, verbose_name='Полное наименование', blank=True, null=True)

    class Meta:
        abstract = True

    @staticmethod
    def get_shortname_map():
        return {
            subclass.get_shortname(): subclass
            for subclass in Hierarchy.__subclasses__()
        }

    @classmethod
    def get_shortname(cls):
        shortname = ''

        if cls.__name__.endswith(Hierarchy.__name__):
            shortname = cls.__name__[:-len(Hierarchy.__name__)]

        return shortname.lower()


class AdmHierarchy(BaseAdmHierarchy, Hierarchy, RegionCodeModelMixin):
    """
    Сведения по иерархии в административном делении
    """
    class Meta:
        verbose_name = verbose_name_plural = 'Иерархия в административном делении'

        indexes = [
            UpperGinIndex(
                fields=['name_with_parents'],
                name='adm_name_with_parents_gin',
                opclasses=['gin_trgm_ops'],
            ),
        ]


class MunHierarchy(BaseMunHierarchy, Hierarchy, RegionCodeModelMixin):
    """
    Сведения по иерархии в муниципальном делении
    """
    class Meta:
        verbose_name = verbose_name_plural = 'Иерархия в муниципальном делении'

        indexes = [
            UpperGinIndex(
                fields=['name_with_parents'],
                name='mun_name_with_parents_gin',
                opclasses=['gin_trgm_ops'],
            ),
        ]

    def get_allowed_oktmo(self):
        return [
            self.oktmo[:2].ljust(8, '0'),  # Федеральный уровень
            self.oktmo[:5].ljust(8, '0'),  # Муниципальный уровень
            self.oktmo[:8],  # Уровень населенного пункта
        ]

    def get_ancestors(self, include_self=False):
        ancestors = super().get_ancestors(include_self)

        already_used_ids = []
        unique_ancestors = []

        for each in ancestors:
            if each.id not in already_used_ids:
                already_used_ids.append(each.id)
                unique_ancestors.append(each)

        if len(ancestors) == len(unique_ancestors):
            result = ancestors
        else:
            # Если попали в эту ветку, значит в иерархии есть несколько
            # родителей с одним id, но разным ОКТМО
            allowed_oktmo = self.get_allowed_oktmo()
            result = []
            for ancestor in unique_ancestors:
                if ancestor.oktmo in allowed_oktmo or (include_self and ancestor.id == self.id):
                    result.append(ancestor)

        return result


make_fk(AdmHierarchy, 'objectid', to=ReestrObjects)
make_fk(MunHierarchy, 'objectid', to=ReestrObjects)

make_fk(AdmHierarchy, 'parentobjid', to=ReestrObjects)
make_fk(MunHierarchy, 'parentobjid', to=ReestrObjects)
